/**
 * Manager is base class for management / group / list of storable objects
 */
class Manager extends KeyLists
{
	constructor() {
		::KeyLists.constructor();
	}
}

class ServiceManager extends Manager
{
	constructor() {
		::Manager.constructor();
	}

	/**
	 * Insert new service item
	 * actually only insert it storage table
	 * @param serv Class of service
	 * @param priority Priority of service
	 * @return a valid ID of servable class
	 */
	function New(serv, priority)
	{
		assert(serv instanceof Services);
        local c = 1;
        while (this.list.HasItem(c)) c++;
        serv.Mem.ID = c;
        this.Push(c, priority, serv.Mem.GetStorage());
        return c;
	}

	/**
	 * Find an ID using key of service
	 * @param key Key to find
	 * @return a valid ID or null if not found
	 */
	function FindKey(key)
	{
		foreach (idx, val in this.list) if (this.Item(idx)["Key"] == key) return idx;
	}
}

class ServableManager extends Manager
{
	constructor()
	{
		::Manager.constructor();
	}
	
	/**
     * Overriden to Get a class from list
     * @param item ID
     * @return clss of item
     */
	function Item(id)
	{
		return Servable.MakeClass(this._storage.rawget(id)); 
	}
	
	/**
	 * Insert New Servable Item.
	 * actually only insert it storage table
	 * @param serv Servable class to insert
	 * @return Available ID
	 */
	 function New(serv)
	 {
	 	assert(serv instanceof Servable);
        local c = 1;
        while (this.list.HasItem(c)) c++;
        local val = serv.GetLocation();
        this.Push(c, val, serv.GetStorage());
        return c;
	 }
	 
	 function FindID(serv)
	 {
	 	assert(serv instanceof Servable);
	 	foreach (idx, val in this.list) {
	 		if (this.GetValue(idx) == serv.GetLocation()) return idx;
	 	}
	 }
}

/**
 * Task management
 */
class TaskManager extends Manager
{
	/** local  ticker */
	tick =  null;

	constructor()
	{
		::Manager.constructor();
		tick = 0;
	}

	/**
	 * Insert new task into scheduler
	 * actually insert class of task item
	 * @param task Class of TaskItem to insert
	 * @param looper In what tick will executed
	 * @return the ID of this task in scheduler
	 */
	function New(task, looper)
	{
		assert(task instanceof TaskItem);
		local c = 1;
		while (this.list.HasItem(c)) c++;
		this.Push(c, looper, task);
		return c;
	}

	/**
	 * Run the scheduler
	 * execute scheduled task by tick
	 */
	function Run () {
		if (this.list.IsEmpty()) return;
		this.tick ++;
		local task = null;
		//Sort the tasks
		this.SortValueAscending();
		// iterating tasks
		for (local i= this.list.Begin(); this.list.HasNext(); i = this.list.Next()) {
			// sleep time each task
			AIController.Sleep(10);
			// get a task class from managers storage
			task = this.Item(i);
			// is it time to run ?
			if (this.tick % this.GetValue(i) == 0) {
				AILog.Info("Run:" + task.GetName());
				task.Execute();
				// check if task is kind of yield task
				if (task instanceof YieldTask) {
					if (task.GetResult() != "stop") continue;
				}
				// only remove if the task is removable and generator died
				if (task.IsRemovable()) this.RemoveItem(i);
			} else {
				AILog.Info("Suspend:" + task.GetName());
			}
			task = null;
		}
		//maximum tick
		if (this.tick == 0xFFFF) this.tick = 0;
	}
}
