/**
 * Base all class that has storage table
 */
class Storage
{
	/// Storage to support save/load
	_storage = null;

	constructor()
	{
		this._storage = {};
	}

	/**
	 * Get this class storage
	 * @return current storage table
	 */
	function GetStorage() { return this._storage; }

	/**
	 * Set class storage
	 * @param val table to set into class
	 */
	function SetStorage( val) { this._storage = val; }

}

/**
 * TransAI Working Memory
 * A class for storing during AI play
 * @note Not recomended to extends this class
 */
class Memory extends Storage
{
	constructor() {
		::Storage.constructor();
	}

	/**
	 * internal _get method
	 * @param idx index of table to get
	 * @return value of index
	 */
	function _get(idx)
	{
		if (idx in this) return this[idx];
		if (idx in this._storage) return this._storage[idx];
		::print("not found in class storage");
	}

	/**
	 * internal _set method
	 * @param idx index of table to set
	 * @param val Value to set
	 */
	function _set(idx, val)
	{
		if((typeof val == "class") || (typeof val == "instance")) throw "using " + typeof val;
		this._storage.rawset(idx, val);
	}
}

/**
 * Storable Object with id
 * A base class for object that has an ID and storage
 */
class Storable extends Storage
{
	constructor() {
		::Storage.constructor();
		this._storage._id <- 0; ///< ID of this object
	}

	/**
	 * Get ID of this object
	 * @return Id of this object
	 */
	function GetID() {return this._storage._id; }

	/**
	 * Set ID of this object
	 * @param val ID to set
	 */
	function SetID(val) {this._storage._id = val; }
}

/**
 * Base class for collection of storable objects
 */
class KeyLists extends Storage
{
	/** Internal List */
	list = null;

	constructor() {
		::Storage.constructor();
		this.list = AIList();
	}

	/**
	 * Count of list
	 * @return Count of storable item has
	 */
	function Count() {return this.list.Count(); }

	/**
	 * Add Item and it storage to list
	 * @param item ID of item
	 * @param storage of item
	 */
	function AddItem(item, storage)
	{
		this.list.AddItem(item, 0);
		this._storage[item] <- storage;
	}

	/**
	 * Change Item and it storage already in the list
	 * @param item ID of item
	 * @param storage of item
	 */
	function ChangeItem(item, storage)
	{
		this._storage[item] = storage;
	}

	/**
	 * Remove Item and it storage in the list
	 * @param item ID of item
	 */
	function RemoveItem(item)
	{
		this.ChangeItem(item, null);
		this._storage.rawdelete(item);
		this.list.RemoveItem(item);
	}

	/**
	 * Find an Item in the list
	 * @param item ID of item
	 * @return true if found
	 */
	function HasItem(item)
	{
		return this.list.HasItem(item) && this._storage.rawin(item);
	}

	/**
	 * Set value of Item (for sorting)
	 * @param item ID
	 * @param value of item
	 */
	function SetValue(item, value) { return this.list.SetValue(item, value); }

	/**
	 * Get value of Item (for sorting)
	 * @param item ID
	 * @param value of item
	 */
	function GetValue(item) { return this.list.GetValue(item); }

	/**
	 * Add Item, value and it storage to list
	 * @param item ID
	 * @param value of item
	 * @param storage of item
	 */
	function Push(item, value, storage)
	{
		if (this.HasItem(item)) this.ChangeItem(item, storage);
		else this.AddItem(item, storage);
		this.SetValue(item, value);
	}

	/**
	 * Get a storage from list
	 * @param item ID
	 * @return storage of item
	 */
	function Item(id) { return this._storage.rawget(id); }

	/**
	 * Sort list ascending by it value
	 */
	function SortValueAscending()
	{
		this.list.Sort(AIAbstractList.SORT_BY_VALUE, true);
	}
	
	/**
	 * Valuate list for sorting
	 * @param valuator function to be used
	 * @param ... list of arguments
	 */
	function Valuate(valuator, ...)
	{
		assert(typeof(valuator) == "function");
		local args = [this, null];
		for(local c = 1; c < vargc; c++) args.push(vargv[c]);
		foreach(idx, val in this.list) {
			args[1] = idx;
			local value = valuator.acall(args);
			if (typeof(value) == "bool") {
				value = value ? 1 : 0;
			} else if (typeof(value) != "integer") throw("Invalid return type from valuator");
			this.SetValue(idx, value);
		}
	}
}
